#include <stdlib.h>
#include <math.h>
#include "Scene.h"

#define PI 3.1415926535897932384626433832795

Scene::Scene()
{
   head_ptr = NULL;
   tail_ptr = NULL;

   textureCount = 0;
   bilerp_flag = false;

   projection = Matrix(1.0); /* Identity matrix */
         view = Matrix(1.0);
        model = Matrix(1.0);

   light = NULL;

   vp_width  = 1000;  /* Viewport width */
   vp_height =  800;  /* Viewport height */
}


Scene::~Scene()
{
   destroyList();
}


void Scene::addTexture( Texture *tex)
{
   textures[textureCount++] = tex;
   return;
}


void Scene::addTriangle( Triangle *tri )
{
   if (head_ptr == NULL) /* insert the first Triangle into an empty list */
   {
      head_ptr = new TriangleListNode();
      head_ptr->t = tri;
      tail_ptr = head_ptr;
   }
   else
   {
      tail_ptr->next = new TriangleListNode();
      tail_ptr = tail_ptr->next;
      tail_ptr->t = tri;
   }
   return;
}


void Scene::proj2Identity()
{
   projection = Matrix(1.0);
   return;
}


void Scene::projMult(Matrix m)
{
   projection = projection.times(m);
   return;
}


void Scene::projPerspective(double fovy, double aspect, double zNear, double zFar)
{
   /*  http://www.opengl.org/sdk/docs/man/xhtml/gluPerspective.xml  */
   double f = 1.0/tan((PI/180.0)*fovy/2.0);
   projection = projection.times(Matrix(Vector(f/aspect, 0.0, 0.0, 0.0),
                                        Vector(0.0, f, 0.0, 0.0),
                                        Vector(0.0, 0.0, (zFar+zNear)/(zNear-zFar), -1.0),
                                        Vector(0.0, 0.0, (2.0*zFar*zNear)/(zNear-zFar), 0.0)));
   return;
}


void Scene::projFrustum(double left, double right, double bottom, double top, double near, double far)
{
   /*  http://www.opengl.org/sdk/docs/man/xhtml/glFrustum.xml  */
   double A = (right+left)/(right-left);
   double B = (top+bottom)/(top-bottom);
   double C =   (far+near)/(far-near);
   double D = (2*far*near)/(far-near);
   projection = projection.times(Matrix(Vector((2.0*near)/(right-left), 0.0, 0.0, 0.0),
                                        Vector(0.0, (2.0*near)/(top-bottom), 0.0, 0.0),
                                        Vector(  A,   B,  C, -1.0),
                                        Vector(0.0, 0.0,  D,  0.0)));
   return;
}


void Scene::projOrtho(double left, double right, double bottom, double top, double near, double far)
{
   /*  http://www.opengl.org/sdk/docs/man/xhtml/glOrtho.xml  */
   double tx = -(right+left)/(right-left);
   double ty = -(top+bottom)/(top-bottom);
   double tz =   -(far+near)/(far-near);
   projection = projection.times(Matrix(Vector((2.0)/(right-left), 0.0, 0.0, 0.0),
                                        Vector(0.0, (2.0)/(top-bottom), 0.0, 0.0),
                                        Vector(0.0, 0.0, (-2.0)/(far-near), 0.0),
                                        Vector(tx, ty, tz, 1.0)));
   return;
}


void Scene::view2Identity()
{
   view = Matrix(1.0);
   return;
}


void Scene::viewMult(Matrix m)
{
   view = view.times(m);
   return;
}


void Scene::viewLookAt(double eyex, double eyey, double eyez, double centerx, double centery, double centerz, double upx, double upy, double upz)
{
   /*  http://www.opengl.org/sdk/docs/man/xhtml/gluLookAt.xml  */
   Vector F = Vector(centerx - eyex, centery - eyey, centerz - eyez);
   Vector UP = Vector(upx, upy, upz);
   Vector f = F.normalize();
   Vector up = UP.normalize();
   Vector s = f.crossProduct(up);
   Vector u = s.crossProduct(f);
   view = view.times(Matrix(Vector(s.x, u.x, -f.x, 0.0),
                            Vector(s.y, u.y, -f.y, 0.0),
                            Vector(s.z, u.z, -f.z, 0.0),
                            Vector(0.0, 0.0,  0.0, 1.0)));

   view = view.times(Matrix(Vector(-eyex, -eyey, -eyez)));
   return;
}


void Scene::viewRotate(double theta, double x, double y, double z)
{
   view = view.times(Matrix(theta, x, y, z));
   return;
}


void Scene::viewScale(double x, double y, double z)
{
   view = view.times(Matrix(x, y, z));
   return;
}


void Scene::viewTranslate(double x, double y, double z)
{
   view = view.times(Matrix(Vector(x, y, z)));
   return;
}



void Scene::model2Identity()
{
   model = Matrix(1.0);
   return;
}


void Scene::modelMult(Matrix m)
{
   model = model.times(m);
   return;
}


void Scene::modelRotate(double theta, double x, double y, double z)
{
   model = model.times(Matrix(theta, x, y, z));
   return;
}


void Scene::modelScale(double x, double y, double z)
{
   model = model.times(Matrix(x, y, z));
   return;
}


void Scene::modelTranslate(double x, double y, double z)
{
   model = model.times(Matrix(Vector(x, y, z)));
   return;
}


void Scene::setViewport(int width, int height)
{
   vp_width  = width;
   vp_height = height;
}


void Scene::destroyList()
{
   TriangleListNode *ptr;

   while (head_ptr)
   {
      ptr = head_ptr->next;

      free(head_ptr->t); /* free the Triangle object */
      free(head_ptr);    /* free the TriangleListNode object */

      head_ptr = ptr;
   }
   head_ptr = NULL;
   tail_ptr = NULL;
   return;
}
